"""
Tests of uploads and downloads using local fakes and requiring no credentials.
"""
import io

import pytest
import httptest
import aiohttp

import taskcluster
from taskcluster import upload, download


class FakeQueue:
    def __init__(self, storageType, ts):
        self.storageType = storageType
        self.ts = ts
        self.options = {"rootUrl": "https://tc-testing.example.com"}

    def latestArtifact(self, taskId, name):
        return self.artifact(taskId, 1, name)

    def artifact(self, taskId, runId, name):
        assert taskId == 'task-id'
        assert runId == 1
        assert name == 'public/test.data'

        if self.storageType == 's3' or self.storageType == 'reference':
            return {
                "storageType": self.storageType,
                "url": f"{self.ts.url()}data",
            }

        elif self.storageType == 'object':
            return {
                "storageType": self.storageType,
                "name": "some/object",
                "credentials": {"clientId": "c", "accessToken": "a"},
            }

        elif self.storageType == 'error':
            return {
                "storageType": self.storageType,
                "message": "uhoh",
                "reason": "testing",
            }


class FakeObject:
    def __init__(self, ts):
        self.ts = ts

    def startDownload(self, name, payload):
        assert payload["acceptDownloadMethods"]["simple"]
        return {
            "method": "simple",
            "url": f"{self.ts.url()}data",
        }

    def createUpload(self, name, payload):
        self.lastUploadId = payload["uploadId"]
        self.lastProjectId = payload["projectId"]

        putUrl = payload["proposedUploadMethods"]["putUrl"]
        assert putUrl["contentType"] == "text/plain"
        return {
            "expires": payload["expires"],
            "projectId": payload["projectId"],
            "uploadId": payload["uploadId"],
            "uploadMethod": {
                "putUrl": {
                    "expires": payload["expires"],
                    "headers": {
                        "content-type": putUrl["contentType"],
                        "content-length": str(putUrl["contentLength"]),
                    },
                    "url": f"{self.ts.url()}data",
                },
            },
        }

    def finishUpload(self, name, payload):
        assert payload["uploadId"] == self.lastUploadId
        assert payload["projectId"] == self.lastProjectId

        return {}


def test_simple_download_fails():
    "When a simple download's GET fails with a 400, an exception is raised and no retries occur"
    getcount = 0

    class Server(httptest.Handler):
        def do_GET(self):
            nonlocal getcount
            getcount += 1
            self.send_response(400)
            self.end_headers()
            self.wfile.write(b'uhoh')

    with httptest.Server(Server) as ts:
        objectService = FakeObject(ts)
        with pytest.raises(aiohttp.ClientResponseError):
            download.downloadToBuf(
                name="some/object",
                objectService=objectService)
        assert getcount == 1


def test_simple_download_fails_retried():
    "When a simple download's GET fails with a 500, an exception is raised after five retries"
    attempts = 0

    class Server(httptest.Handler):
        def do_GET(self):
            nonlocal attempts
            attempts += 1
            self.send_response(500)
            self.end_headers()
            self.wfile.write(b'uhoh')

    with httptest.Server(Server) as ts:
        objectService = FakeObject(ts)
        with pytest.raises(aiohttp.ClientResponseError):
            download.downloadToBuf(
                name="some/object",
                objectService=objectService)

    assert attempts == 6  # one try plus five retries


def test_simple_download_fails_retried_succeeds(randbytes):
    "When a simple download's GET fails with a 500, it is retried successfully"
    attempts = 0
    data = randbytes(1024)

    class Server(httptest.Handler):
        def do_GET(self):
            nonlocal attempts
            attempts += 1
            if attempts > 2:
                self.send_response(200)
                self.send_header('content-type', 'text/plain')
                self.send_header('content-length', str(len(data)))
                self.end_headers()
                self.wfile.write(data)
            else:
                self.send_response(500)
                self.end_headers()
                self.wfile.write(b'uhoh')

    with httptest.Server(Server) as ts:
        objectService = FakeObject(ts)
        buf, content_type = download.downloadToBuf(
            name="some/object",
            objectService=objectService)

    assert attempts == 3
    assert buf == data
    assert content_type == 'text/plain'


def test_download(randbytes):
    """Test that download works with a custom sync writer factory."""
    data = randbytes(1024)

    class Server(httptest.Handler):
        def do_GET(self):
            self.send_response(200)
            self.end_headers()
            self.wfile.write(data)

    writer = None

    def writerFactory():
        nonlocal writer
        writer = io.BytesIO()
        return writer

    with httptest.Server(Server) as ts:
        objectService = FakeObject(ts)
        download.download(
            name="some/object",
            writerFactory=writerFactory,
            objectService=objectService)

    assert bytes(writer.getbuffer()) == data


def test_putUrl_upload_fails(randbytes):
    "When a putUrl upload's PUT fails with a 400, an exception is raised"
    data = randbytes(10240)  # >8k to avoid using dataInline
    attempts = 0

    class Server(httptest.Handler):
        def do_PUT(self):
            nonlocal attempts
            attempts += 1
            self.rfile.read(len(data))
            self.send_response(400)
            self.end_headers()
            self.wfile.write(b'uhoh')

    with httptest.Server(Server) as ts:
        objectService = FakeObject(ts)
        with pytest.raises(aiohttp.ClientResponseError):
            upload.uploadFromBuf(
                projectId="taskcluster",
                expires=taskcluster.fromNow('1 hour'),
                contentType="text/plain",
                contentLength=len(data),
                name="some/object",
                data=data,
                objectService=objectService)

    assert attempts == 1


def test_putUrl_upload_fails_retried(randbytes):
    "When a putUrl upload's PUT fails with a 500, an exception is raised"
    data = randbytes(10240)  # >8k to avoid using dataInline
    attempts = 0

    class Server(httptest.Handler):
        def do_PUT(self):
            nonlocal attempts
            attempts += 1
            self.rfile.read(len(data))
            self.send_response(500)
            self.end_headers()
            self.wfile.write(b'uhoh')

    with httptest.Server(Server) as ts:
        objectService = FakeObject(ts)
        with pytest.raises(aiohttp.ClientResponseError):
            upload.uploadFromBuf(
                projectId="taskcluster",
                expires=taskcluster.fromNow('1 hour'),
                contentType="text/plain",
                contentLength=len(data),
                name="some/object",
                data=data,
                objectService=objectService)

    assert attempts == 6  # one try plus five retries


def test_putUrl_upload_fails_retried_succeeds(randbytes):
    "When a putUrl upload's PUT fails with a 500, an exception is raised"
    data = randbytes(10240)  # >8k to avoid using dataInline
    attempts = 0

    class Server(httptest.Handler):
        def do_PUT(self):
            nonlocal attempts
            attempts += 1
            self.rfile.read(len(data))
            if attempts > 2:
                self.send_response(200)
                self.send_header('content-length', "0")
                self.end_headers()
            else:
                self.send_response(500)
                self.send_header('content-length', "4")
                self.end_headers()
                self.wfile.write(b'uhoh')

    with httptest.Server(Server) as ts:
        objectService = FakeObject(ts)
        upload.uploadFromBuf(
            projectId="taskcluster",
            expires=taskcluster.fromNow('1 hour'),
            contentType="text/plain",
            contentLength=len(data),
            name="some/object",
            data=data,
            objectService=objectService)

    assert attempts == 3


def test_putUrl_upload(randbytes):
    """Test that upload works with a custom sync reader factory."""
    data = randbytes(10240)  # >8k to avoid using dataInline

    uploaded_data = b''

    class Server(httptest.Handler):
        def do_PUT(self):
            nonlocal uploaded_data
            uploaded_data = self.rfile.read(len(data))
            self.send_response(200)
            self.end_headers()

    def readerFactory():
        return io.BytesIO(data)

    with httptest.Server(Server) as ts:
        objectService = FakeObject(ts)
        upload.upload(
            projectId="taskcluster",
            expires=taskcluster.fromNow('1 hour'),
            contentType="text/plain",
            contentLength=len(data),
            name="some/object",
            readerFactory=readerFactory,
            objectService=objectService)

    assert uploaded_data == data


def test_download_object_artifact(randbytes, monkeypatch):
    "Download an object artifact"
    # note: other artifact types are tested in tests/aio; this just
    # validates the sync wrappers
    data = randbytes(1024)

    class Server(httptest.Handler):
        def do_GET(self):
            self.send_response(200)
            self.send_header('content-type', 'text/plain')
            self.send_header('content-length', str(len(data)))
            self.end_headers()
            self.wfile.write(data)

    with httptest.Server(Server) as ts:
        # the wrapped async download will create an _async_ Object client,
        # so we use the fake from the async tests
        def make_fake_async_object(options):
            from aio.test_upload_download import FakeObject
            assert options["credentials"] == {"clientId": "c", "accessToken": "a"}
            assert options["rootUrl"] == "https://tc-testing.example.com"
            return FakeObject(ts)

        monkeypatch.setattr(taskcluster.aio.download, "Object", make_fake_async_object)

        queueService = FakeQueue("object", ts)
        buf, content_type = download.downloadArtifactToBuf(
            taskId='task-id',
            runId=1,
            name="public/test.data",
            queueService=queueService)

    assert buf == data
    assert content_type == 'text/plain'



"""
Support for uploading objects to the object service, following best
practices for that service.

Data for upload is read from a "reader" provided by a "reader factory".  A
reader has a `read(max_size)` method which reads and returns a chunk of 1 ..
`max_size` bytes, or returns an empty string at EOF.  A reader factory is a
callable which returns a fresh reader, ready to read the first byte of the
object.  When uploads are retried, the reader factory may be called more than
once.

This module provides several pre-defined readers and reader factories for
common cases.
"""
import functools

from .aio import upload as aio_upload
from .aio.asyncutils import ensureCoro, runAsync


DATA_INLINE_MAX_SIZE = 8192


def uploadFromBuf(*, data, **kwargs):
    """
    Convenience method to upload data from an in-memory buffer.  Arguments are the same
    as `upload` except that `readerFactory` should not be supplied.
    """
    return runAsync(aio_upload.uploadFromBuf(data=data, **kwargs))


def uploadFromFile(*, file, **kwargs):
    """
    Convenience method to upload data from a a file.  The file should be open
    for reading, in binary mode, and be seekable (`f.seek`).  Remaining
    arguments are the same as `upload` except that `readerFactory` should not
    be supplied.
    """
    return runAsync(aio_upload.uploadFromFile(file=file, **kwargs))


def upload(*, readerFactory, **kwargs):
    """
    Upload the given data to the object service with the given metadata.
    The `maxRetries` parameter has the same meaning as for service clients.
    The `objectService` parameter is an instance of the Object class,
    configured with credentials for the upload.
    """
    wrappedReaderFactory = _wrapSyncReaderFactory(readerFactory)
    return runAsync(aio_upload.upload(readerFactory=wrappedReaderFactory, **kwargs))


def _wrapSyncReaderFactory(readerFactory):
    """Modify the reader returned by readerFactory to have an async read."""
    @functools.wraps(readerFactory)
    async def wrappedFactory():
        reader = readerFactory()
        reader.read = ensureCoro(reader.read)
        return reader

    return wrappedFactory



"""
Support for downloading objects from the object service, following best
practices for that service.

Downloaded data is written to a "writer" provided by a "writer factory".  A
writer has a `write` method which writes the entire passed buffer to storage.
A writer factory is a callable which returns a fresh writer, ready to write the
first byte of the object.  When downloads are retried, the writer factory may
be called more than once.

This module provides several pre-defined writers and writer factories for
common cases.
"""
import functools

from .aio import download as aio_download
from .aio.asyncutils import ensureCoro, runAsync


def downloadToBuf(**kwargs):
    """
    Convenience method to download data to an in-memory buffer and return the
    downloaded data.  Arguments are the same as `download`, except that
    `writerFactory` should not be supplied.  Returns a tuple (buffer, contentType).
    """
    return runAsync(aio_download.downloadToBuf(**kwargs))


def downloadToFile(file, **kwargs):
    """
    Convenience method to download data to a file object.  The file must be
    writeable, in binary mode, seekable (`f.seek`), and truncatable
    (`f.truncate`) to support retries.  Arguments are the same as `download`,
    except that `writerFactory` should not be supplied.  Returns the content-type.
    """
    return runAsync(aio_download.downloadToFile(file=file, **kwargs))


def download(*, writerFactory, **kwargs):
    """
    Download the named object from the object service, using a writer returned
    from `writerFactory` to write the data.  The `maxRetries` parameter has
    the same meaning as for service clients.  The `objectService` parameter is
    an instance of the Object class, configured with credentials for the
    upload.  Returns the content-type.
    """
    wrappedWriterFactory = _wrapSyncWriterFactory(writerFactory)
    return runAsync(aio_download.download(writerFactory=wrappedWriterFactory, **kwargs))


def downloadArtifactToBuf(**kwargs):
    """
    Convenience method to download an artifact to an in-memory buffer and return the
    downloaded data.  Arguments are the same as `downloadArtifact`, except that
    `writerFactory` should not be supplied.  Returns a tuple (buffer, contentType).
    """
    return runAsync(aio_download.downloadArtifactToBuf(**kwargs))


def downloadArtifactToFile(file, **kwargs):
    """
    Convenience method to download an artifact to a file object.  The file must be
    writeable, in binary mode, seekable (`f.seek`), and truncatable
    (`f.truncate`) to support retries.  Arguments are the same as `downloadArtifac`,
    except that `writerFactory` should not be supplied.  Returns the content-type.
    """
    return runAsync(aio_download.downloadArtifactToFile(file=file, **kwargs))


def downloadArtifact(*, writerFactory, **kwargs):
    """
    Download the named artifact with the appropriate storageType, using a writer returned
    from `writerFactory` to write the data.  The `maxRetries` parameter has
    the same meaning as for service clients.  The `queueService` parameter is
    an instance of the Queue class, configured with credentials for the
    download.  Returns the content-type.
    """
    wrappedWriterFactory = _wrapSyncWriterFactory(writerFactory)
    return runAsync(aio_download.downloadArtifact(writerFactory=wrappedWriterFactory, **kwargs))


def _wrapSyncWriterFactory(writerFactory):
    """Modify the reader returned by readerFactory to have an async read."""
    @functools.wraps(writerFactory)
    async def wrappedFactory():
        writer = writerFactory()
        writer.write = ensureCoro(writer.write)
        return writer

    return wrappedFactory



"""
Support for downloading objects from the object service, following best
practices for that service.

Downloaded data is written to a "writer" provided by a "writer factory".  A
writer has an async `write` method which writes the entire passed buffer to
storage.  A writer factory is an async callable which returns a fresh writer,
ready to write the first byte of the object.  When downloads are retried, the
writer factory may be called more than once.

Note that `aiofile.open` returns a value suitable for use as a writer, if async
file IO is important to the application.

This module provides several pre-defined writers and writer factories for
common cases.
"""
import aiohttp
import contextlib

from .asyncutils import ensureCoro
from .reader_writer import streamingCopy, BufferWriter, FileWriter
from .retry import retry
from . import Object
from ..exceptions import TaskclusterArtifactError, TaskclusterFailure


async def downloadToBuf(**kwargs):
    """
    Convenience method to download data to an in-memory buffer and return the
    downloaded data.  Arguments are the same as `download`, except that
    `writerFactory` should not be supplied.  Returns a tuple (buffer, contentType).
    """
    writer = None

    async def writerFactory():
        nonlocal writer
        writer = BufferWriter()
        return writer

    contentType = await download(writerFactory=writerFactory, **kwargs)
    return writer.getbuffer(), contentType


async def downloadToFile(file, **kwargs):
    """
    Convenience method to download data to a file object.  The file must be
    writeable, in binary mode, seekable (`f.seek`), and truncatable
    (`f.truncate`) to support retries.  Arguments are the same as `download`,
    except that `writerFactory` should not be supplied.  Returns the content-type.
    """
    async def writerFactory():
        file.seek(0)
        file.truncate()
        return FileWriter(file)

    return await download(writerFactory=writerFactory, **kwargs)


async def download(*, name, maxRetries=5, objectService, writerFactory):
    """
    Download the named object from the object service, using a writer returned
    from `writerFactory` to write the data.  The `maxRetries` parameter has
    the same meaning as for service clients.  The `objectService` parameter is
    an instance of the Object class, configured with credentials for the
    download.  Returns the content-type.
    """
    async with aiohttp.ClientSession() as session:
        downloadResp = await ensureCoro(objectService.startDownload)(name, {
            "acceptDownloadMethods": {
                "simple": True,
            },
        })

        method = downloadResp["method"]

        if method == "simple":
            async def tryDownload(retryFor):
                with _maybeRetryHttpRequest(retryFor):
                    writer = await writerFactory()
                    url = downloadResp['url']
                    return await _doSimpleDownload(url, writer, session)

            return await retry(maxRetries, tryDownload)
        else:
            raise RuntimeError(f'Unknown download method {method}')


async def downloadArtifactToBuf(**kwargs):
    """
    Convenience method to download an artifact to an in-memory buffer and return the
    downloaded data.  Arguments are the same as `downloadArtifact`, except that
    `writerFactory` should not be supplied.  Returns a tuple (buffer, contentType).
    """
    writer = None

    async def writerFactory():
        nonlocal writer
        writer = BufferWriter()
        return writer

    contentType = await downloadArtifact(writerFactory=writerFactory, **kwargs)
    return writer.getbuffer(), contentType


async def downloadArtifactToFile(file, **kwargs):
    """
    Convenience method to download an artifact to a file object.  The file must be
    writeable, in binary mode, seekable (`f.seek`), and truncatable
    (`f.truncate`) to support retries.  Arguments are the same as `downloadArtifac`,
    except that `writerFactory` should not be supplied.  Returns the content-type.
    """
    async def writerFactory():
        file.seek(0)
        file.truncate()
        return FileWriter(file)

    return await downloadArtifact(writerFactory=writerFactory, **kwargs)


async def downloadArtifact(*, taskId, name, runId=None, maxRetries=5, queueService, writerFactory):
    """
    Download the named artifact with the appropriate storageType, using a writer returned
    from `writerFactory` to write the data.  The `maxRetries` parameter has
    the same meaning as for service clients.  The `queueService` parameter is
    an instance of the Queue class, configured with credentials for the
    download.  Returns the content-type.
    """
    if runId is None:
        artifact = await ensureCoro(queueService.latestArtifact)(taskId, name)
    else:
        artifact = await ensureCoro(queueService.artifact)(taskId, runId, name)

    if artifact["storageType"] == 's3' or artifact["storageType"] == 'reference':
        async with aiohttp.ClientSession() as session:

            async def tryDownload(retryFor):
                with _maybeRetryHttpRequest(retryFor):
                    writer = await writerFactory()
                    return await _doSimpleDownload(artifact["url"], writer, session)

            return await retry(maxRetries, tryDownload)

    elif artifact["storageType"] == 'object':
        objectService = Object({
            "rootUrl": queueService.options["rootUrl"],
            "maxRetries": maxRetries,
            "credentials": artifact["credentials"],
        })
        return await download(
            name=artifact["name"],
            maxRetries=maxRetries,
            objectService=objectService,
            writerFactory=writerFactory)

    elif artifact["storageType"] == 'error':
        raise TaskclusterArtifactError(artifact["message"], artifact["reason"])

    else:
        raise TaskclusterFailure(f"Unknown storageType f{artifact['storageType']}")


@contextlib.contextmanager
def _maybeRetryHttpRequest(retryFor):
    "Catch errors from an aiohttp request and retry the retriable responses."
    try:
        yield
    except aiohttp.ClientResponseError as exc:
        # treat 4xx's as fatal, and retry others
        if 400 <= exc.status < 500:
            raise exc
        return retryFor(exc)
    except aiohttp.ClientError as exc:
        # retry for all other aiohttp errors
        return retryFor(exc)
    # .. anything else is considered fatal


async def _doSimpleDownload(url, writer, session):
    async with session.get(url) as resp:
        contentType = resp.content_type
        resp.raise_for_status()
        # note that `resp.content` is a StreamReader and satisfies the
        # requirements of a reader in this case
        await streamingCopy(resp.content, writer)

    return contentType



"""
Support for uploading objects to the object service, following best
practices for that service.

Data for upload is read from a "reader" provided by a "reader factory".  A
reader has an async `read(max_size)` method which reads and returns a chunk of
1 .. `max_size` bytes, or returns an empty string at EOF.  A reader factory is an async
callable which returns a fresh reader, ready to read the first byte of the
object.  When uploads are retried, the reader factory may be called more than
once.

Note that `aiofile.open` returns a value suitable for use as a reader, if async
file IO is important to the application.

This module provides several pre-defined readers and reader factories for
common cases.
"""
import base64
import hashlib

import aiohttp

import taskcluster
from .asyncutils import ensureCoro
from .reader_writer import streamingCopy, BufferReader, BufferWriter, FileReader
from .retry import retry

DATA_INLINE_MAX_SIZE = 8192


async def uploadFromBuf(*, data, **kwargs):
    """
    Convenience method to upload data from an in-memory buffer.  Arguments are the same
    as `upload` except that `readerFactory` should not be supplied.
    """
    async def readerFactory():
        return BufferReader(data)

    await upload(**kwargs, readerFactory=readerFactory)


async def uploadFromFile(*, file, **kwargs):
    """
    Convenience method to upload data from a a file.  The file should be open
    for reading, in binary mode, and be seekable (`f.seek`).  Remaining
    arguments are the same as `upload` except that `readerFactory` should not
    be supplied.
    """
    async def readerFactory():
        file.seek(0)
        return FileReader(file)

    await upload(**kwargs, readerFactory=readerFactory)


async def upload(*, projectId, name, contentType, contentLength, expires,
                 readerFactory, maxRetries=5, uploadId=None, objectService):
    """
    Upload the given data to the object service with the given metadata.
    The `maxRetries` parameter has the same meaning as for service clients.
    The `objectService` parameter is an instance of the Object class,
    configured with credentials for the upload.
    """
    # wrap the readerFactory with one that will also hash the data
    hashingReader = None

    async def hashingReaderFactory():
        nonlocal hashingReader
        hashingReader = HashingReader(await readerFactory())
        return hashingReader

    async with aiohttp.ClientSession() as session:
        if not uploadId:
            uploadId = taskcluster.slugid.nice()
        proposedUploadMethods = {}

        if contentLength < DATA_INLINE_MAX_SIZE:
            reader = await hashingReaderFactory()
            writer = BufferWriter()
            await streamingCopy(reader, writer)
            encoded = base64.b64encode(writer.getbuffer())
            proposedUploadMethods['dataInline'] = {
                "contentType": contentType,
                "objectData": encoded,
            }

        proposedUploadMethods['putUrl'] = {
            "contentType": contentType,
            "contentLength": contentLength,
        }

        uploadResp = await ensureCoro(objectService.createUpload)(name, {
            "expires": expires,
            "projectId": projectId,
            "uploadId": uploadId,
            "proposedUploadMethods": proposedUploadMethods,
        })

        async def tryUpload(retryFor):
            try:
                uploadMethod = uploadResp["uploadMethod"]
                if 'dataInline' in uploadMethod:
                    # data is already uploaded -- nothing to do
                    pass
                elif 'putUrl' in uploadMethod:
                    reader = await hashingReaderFactory()
                    await _putUrlUpload(uploadMethod['putUrl'], reader, session)
                else:
                    raise RuntimeError("Could not negotiate an upload method")
            except aiohttp.ClientResponseError as exc:
                # treat 4xx's as fatal, and retry others
                if 400 <= exc.status < 500:
                    raise exc
                return retryFor(exc)
            except aiohttp.ClientError as exc:
                # retry for all other aiohttp errors
                return retryFor(exc)
            # .. anything else is considered fatal

        await retry(maxRetries, tryUpload)

        hashes = hashingReader.hashes(contentLength)

        await ensureCoro(objectService.finishUpload)(name, {
            "projectId": projectId,
            "uploadId": uploadId,
            "hashes": hashes,
        })


async def _putUrlUpload(method, reader, session):
    chunk_size = 64 * 1024

    async def reader_gen():
        while True:
            chunk = await reader.read(chunk_size)
            if not chunk:
                break
            yield chunk

    resp = await session.put(method['url'], headers=method['headers'], data=reader_gen())
    resp.raise_for_status()


class HashingReader:
    """A Reader implementation that hashes contents as they are read."""

    def __init__(self, inner):
        self.inner = inner
        self.sha256 = hashlib.sha256()
        self.sha512 = hashlib.sha512()
        self.bytes = 0

    async def read(self, max_size):
        chunk = await self.inner.read(max_size)
        self.update(chunk)
        return chunk

    def update(self, chunk):
        self.sha256.update(chunk)
        self.sha512.update(chunk)
        self.bytes += len(chunk)

    def hashes(self, contentLength):
        """Return the hashes in a format suitable for finishUpload, first checking that all the bytes
        in the content were hashed."""
        if contentLength != self.bytes:
            raise RuntimeError(f"hashed {self.bytes} bytes but content length is {contentLength}")
        return {
            "sha256": self.sha256.hexdigest(),
            "sha512": self.sha512.hexdigest(),
        }



"""
Utilities supporting the "reader" and "writer" definitions used in uploads and downloads.
"""
import asyncio
import io


class BufferWriter:
    """A writer that writes to an in-memory buffer"""
    def __init__(self):
        self.buf = io.BytesIO()

    async def write(self, chunk):
        self.buf.write(chunk)

    def getbuffer(self):
        """Get the content of the in-memory buffer"""
        return self.buf.getbuffer()


class BufferReader:
    """A reader that reads from an in-memory buffer"""
    def __init__(self, data):
        self.buf = io.BytesIO(data)

    async def read(self, max_size):
        return self.buf.read(max_size)


class FileWriter:
    """A writer that writes to a (sync) file.  The file should be opened in binary mode
    and empty."""
    def __init__(self, file):
        self.file = file

    async def write(self, chunk):
        self.file.write(chunk)


class FileReader:
    """A reader that reads from a (sync) file.  The file should be opened in binary mode,
    and positioned at its beginning."""
    def __init__(self, file):
        self.file = file

    async def read(self, max_size):
        return self.file.read(max_size)


async def streamingCopy(reader, writer):
    "Copy data from a reader to a writer, as those are defined in upload.py and download.py"
    # we will read and write concurrently, but with limited buffering -- just enough
    # that read and write operations are not forced to alternate
    chunk_size = 64 * 1024
    q = asyncio.Queue(maxsize=1)

    async def read_loop():
        while True:
            chunk = await reader.read(chunk_size)
            await q.put(chunk)
            if not chunk:
                break

    async def write_loop():
        while True:
            chunk = await q.get()
            if not chunk:
                q.task_done()
                break
            await writer.write(chunk)
            q.task_done()

    read_task = asyncio.ensure_future(read_loop())
    write_task = asyncio.ensure_future(write_loop())

    try:
        await asyncio.gather(read_task, write_task)
    finally:
        # cancel any still-running tasks, as in case of an exception
        read_task.cancel()
        write_task.cancel()



import logging
import asyncio

from .. import utils

log = logging.getLogger(__name__)


async def retry(maxRetries, tryFn):
    """
    Retry async `tryFn` based on `maxRetries`.  Each call to `tryFn` will pass a callable
    which should be called with the exception object when an exception can be retried.
    Exceptions raised from `tryFn` are treated as fatal.
    """

    retry = -1  # we plus first in the loop, and attempt 1 is retry 0
    while True:
        retry += 1

        # if this isn't the first retry then we sleep
        if retry > 0:
            snooze = float(retry * retry) / 10.0
            log.info('Sleeping %0.2f seconds for exponential backoff', snooze)
            await asyncio.sleep(utils.calculateSleepTime(retry))

        retriableException = None

        def retryFor(exc):
            nonlocal retriableException
            retriableException = exc

        res = await tryFn(retryFor)

        if not retriableException:
            return res

        if retry < maxRetries:
            log.warning(f'Retrying because of: {retriableException}')
            continue

        raise retriableException



# coding=utf-8
#####################################################
# THIS FILE IS AUTOMATICALLY GENERATED. DO NOT EDIT #
#####################################################
# noqa: E128,E201
from ...aio.asyncclient import AsyncBaseClient
from ...aio.asyncclient import createApiClient
from ...aio.asyncclient import config
from ...aio.asyncclient import createTemporaryCredentials
from ...aio.asyncclient import createSession
_defaultConfig = config


class PurgeCache(AsyncBaseClient):
    """
    The purge-cache service is responsible for tracking cache-purge requests.

    User create purge requests for specific caches on specific workers, and
    these requests are timestamped.  Workers consult the service before
    starting a new task, and purge any caches older than the timestamp.
    """

    classOptions = {
    }
    serviceName = 'purge-cache'
    apiVersion = 'v1'

    async def ping(self, *args, **kwargs):
        """
        Ping Server

        Respond without doing anything.
        This endpoint is used to check that the service is up.

        This method is ``stable``
        """

        return await self._makeApiCall(self.funcinfo["ping"], *args, **kwargs)

    async def lbheartbeat(self, *args, **kwargs):
        """
        Load Balancer Heartbeat

        Respond without doing anything.
        This endpoint is used to check that the service is up.

        This method is ``stable``
        """

        return await self._makeApiCall(self.funcinfo["lbheartbeat"], *args, **kwargs)

    async def version(self, *args, **kwargs):
        """
        Taskcluster Version

        Respond with the JSON version object.
        https://github.com/mozilla-services/Dockerflow/blob/main/docs/version_object.md

        This method is ``stable``
        """

        return await self._makeApiCall(self.funcinfo["version"], *args, **kwargs)

    async def purgeCache(self, *args, **kwargs):
        """
        Purge Worker Cache

        Publish a request to purge caches named `cacheName` with
        on `workerPoolId` workers.

        If such a request already exists, its `before` timestamp is updated to
        the current time.

        This method is ``stable``
        """

        return await self._makeApiCall(self.funcinfo["purgeCache"], *args, **kwargs)

    async def allPurgeRequests(self, *args, **kwargs):
        """
        All Open Purge Requests

        View all active purge requests.

        This is useful mostly for administors to view
        the set of open purge requests. It should not
        be used by workers. They should use the purgeRequests
        endpoint that is specific to their workerType and
        provisionerId.

        This method is ``stable``
        """

        return await self._makeApiCall(self.funcinfo["allPurgeRequests"], *args, **kwargs)

    async def purgeRequests(self, *args, **kwargs):
        """
        Open Purge Requests for a worker pool

        List the caches for this `workerPoolId` that should to be
        purged if they are from before the time given in the response.

        This is intended to be used by workers to determine which caches to purge.

        This method is ``stable``
        """

        return await self._makeApiCall(self.funcinfo["purgeRequests"], *args, **kwargs)

    async def heartbeat(self, *args, **kwargs):
        """
        Heartbeat

        Respond with a service heartbeat.

        This endpoint is used to check on backing services this service
        depends on.

        This method is ``stable``
        """

        return await self._makeApiCall(self.funcinfo["heartbeat"], *args, **kwargs)

    funcinfo = {
        "allPurgeRequests": {
            'args': [],
            'method': 'get',
            'name': 'allPurgeRequests',
            'output': 'v1/all-purge-cache-request-list.json#',
            'query': ['continuationToken', 'limit'],
            'route': '/purge-cache/list',
            'stability': 'stable',
        },
        "heartbeat": {
            'args': [],
            'method': 'get',
            'name': 'heartbeat',
            'route': '/__heartbeat__',
            'stability': 'stable',
        },
        "lbheartbeat": {
            'args': [],
            'method': 'get',
            'name': 'lbheartbeat',
            'route': '/__lbheartbeat__',
            'stability': 'stable',
        },
        "ping": {
            'args': [],
            'method': 'get',
            'name': 'ping',
            'route': '/ping',
            'stability': 'stable',
        },
        "purgeCache": {
            'args': ['workerPoolId'],
            'input': 'v1/purge-cache-request.json#',
            'method': 'post',
            'name': 'purgeCache',
            'route': '/purge-cache/<workerPoolId>',
            'stability': 'stable',
        },
        "purgeRequests": {
            'args': ['workerPoolId'],
            'method': 'get',
            'name': 'purgeRequests',
            'output': 'v1/purge-cache-request-list.json#',
            'query': ['since'],
            'route': '/purge-cache/<workerPoolId>',
            'stability': 'stable',
        },
        "version": {
            'args': [],
            'method': 'get',
            'name': 'version',
            'route': '/__version__',
            'stability': 'stable',
        },
    }


__all__ = ['createTemporaryCredentials', 'config', '_defaultConfig', 'createApiClient', 'createSession', 'PurgeCache']



"""
Tests of uploads and downloads against the real Object service as given by
TASKCLUSTER_* variables
"""

import os
import secrets

import pytest

import taskcluster
from taskcluster import Object, upload, download
import base


def shouldSkip():
    vars = ['TASKCLUSTER_' + x for x in ('ROOT_URL', 'CLIENT_ID', 'ACCESS_TOKEN')]
    if all(v in os.environ for v in vars):
        return False

    if 'NO_SKIP_TESTS' in os.environ:
        raise RuntimeError('NO_SKIP_TESTS is set but TASKCLUSTER_{ROOT_URL,CLIENT_ID,ACCESS_TOKEN} are not')

    return True


pytestmark = [
    pytest.mark.skipif(shouldSkip(), reason="Skipping tests that require real credentials"),
]


@pytest.fixture
def objectService():
    """
    Return a client object built with TC credentials from the environment.  The client
    is configured with authorized_scopes containing the scopes required for these
    tests.  If creating a new client for these tests, consult the scopes in the
    function body.
    """
    return Object({
        'rootUrl': base.REAL_ROOT_URL,
        'credentials': {
            'clientId': os.environ['TASKCLUSTER_CLIENT_ID'],
            'accessToken': os.environ['TASKCLUSTER_ACCESS_TOKEN'],
        },
        'authorizedScopes': [
            "object:upload:taskcluster:taskcluster/test/client-py/*",
            "object:download:taskcluster/test/client-py/*",
        ],
    })


def test_small_upload_download(objectService):
    """We can upload a small amount of literal data"""
    data = b"hello, world"
    do_over_wire_buf_test(data, objectService)


def test_large_upload_download(objectService):
    """We can upload a larger amount of literal data (>8k to avoid using dataInline)"""
    data = secrets.token_bytes(102400)
    do_over_wire_buf_test(data, objectService)


def do_over_wire_buf_test(data, objectService):
    """We can upload a small amount of literal data"""
    name = f"taskcluster/test/client-py/{taskcluster.slugid.v4()}"
    upload.uploadFromBuf(
        projectId="taskcluster",
        name=name,
        contentType="text/plain",
        contentLength=len(data),
        expires=taskcluster.fromNow('1 hour'),
        data=data,
        objectService=objectService)

    got, contentType = download.downloadToBuf(
        name=name,
        objectService=objectService)

    assert got == data
    assert contentType == 'text/plain'


def test_file_upload_download(objectService, tmp_path):
    src = tmp_path / "src"
    dest = tmp_path / "dest"

    data = secrets.token_bytes(102400)
    with open(src, "wb") as f:
        f.write(data)

    name = f"taskcluster/test/client-py/{taskcluster.slugid.v4()}"
    with open(src, "rb") as file:
        upload.uploadFromFile(
            projectId="taskcluster",
            name=name,
            contentType="text/plain",
            contentLength=len(data),
            expires=taskcluster.fromNow('1 hour'),
            file=file,
            objectService=objectService)

    with open(dest, "wb") as file:
        contentType = download.downloadToFile(
            name=name,
            file=file,
            objectService=objectService)

    with open(dest, "rb") as f:
        got = f.read()

    assert got == data
    assert contentType == 'text/plain'



# coding=utf-8
#####################################################
# THIS FILE IS AUTOMATICALLY GENERATED. DO NOT EDIT #
#####################################################
# noqa: E128,E201
from ...aio.asyncclient import AsyncBaseClient
from ...aio.asyncclient import createApiClient
from ...aio.asyncclient import config
from ...aio.asyncclient import createTemporaryCredentials
from ...aio.asyncclient import createSession
_defaultConfig = config


class Notify(AsyncBaseClient):
    """
    The notification service listens for tasks with associated notifications
    and handles requests to send emails and post pulse messages.
    """

    classOptions = {
    }
    serviceName = 'notify'
    apiVersion = 'v1'

    async def ping(self, *args, **kwargs):
        """
        Ping Server

        Respond without doing anything.
        This endpoint is used to check that the service is up.

        This method is ``stable``
        """

        return await self._makeApiCall(self.funcinfo["ping"], *args, **kwargs)

    async def lbheartbeat(self, *args, **kwargs):
        """
        Load Balancer Heartbeat

        Respond without doing anything.
        This endpoint is used to check that the service is up.

        This method is ``stable``
        """

        return await self._makeApiCall(self.funcinfo["lbheartbeat"], *args, **kwargs)

    async def version(self, *args, **kwargs):
        """
        Taskcluster Version

        Respond with the JSON version object.
        https://github.com/mozilla-services/Dockerflow/blob/main/docs/version_object.md

        This method is ``stable``
        """

        return await self._makeApiCall(self.funcinfo["version"], *args, **kwargs)

    async def email(self, *args, **kwargs):
        """
        Send an Email

        Send an email to `address`. The content is markdown and will be rendered
        to HTML, but both the HTML and raw markdown text will be sent in the
        email. If a link is included, it will be rendered to a nice button in the
        HTML version of the email

        This method is ``experimental``
        """

        return await self._makeApiCall(self.funcinfo["email"], *args, **kwargs)

    async def pulse(self, *args, **kwargs):
        """
        Publish a Pulse Message

        Publish a message on pulse with the given `routingKey`.

        This method is ``experimental``
        """

        return await self._makeApiCall(self.funcinfo["pulse"], *args, **kwargs)

    async def matrix(self, *args, **kwargs):
        """
        Post Matrix Message

        Post a message to a room in Matrix. Optionally includes formatted message.

        The `roomId` in the scopes is a fully formed `roomId` with leading `!` such
        as `!foo:bar.com`.

        Note that the matrix client used by taskcluster must be invited to a room before
        it can post there!

        This method is ``experimental``
        """

        return await self._makeApiCall(self.funcinfo["matrix"], *args, **kwargs)

    async def slack(self, *args, **kwargs):
        """
        Post Slack Message

        Post a message to a Slack channel.

        The `channelId` in the scopes is a Slack channel ID, starting with a capital C.

        The Slack app can post into public channels by default but will need to be added
        to private channels before it can post messages there.

        This method is ``experimental``
        """

        return await self._makeApiCall(self.funcinfo["slack"], *args, **kwargs)

    async def addDenylistAddress(self, *args, **kwargs):
        """
        Denylist Given Address

        Add the given address to the notification denylist. Addresses in the denylist will be ignored
        by the notification service.

        This method is ``experimental``
        """

        return await self._makeApiCall(self.funcinfo["addDenylistAddress"], *args, **kwargs)

    async def deleteDenylistAddress(self, *args, **kwargs):
        """
        Delete Denylisted Address

        Delete the specified address from the notification denylist.

        This method is ``experimental``
        """

        return await self._makeApiCall(self.funcinfo["deleteDenylistAddress"], *args, **kwargs)

    async def listDenylist(self, *args, **kwargs):
        """
        List Denylisted Notifications

        Lists all the denylisted addresses.

        By default this end-point will try to return up to 1000 addresses in one
        request. But it **may return less**, even if more tasks are available.
        It may also return a `continuationToken` even though there are no more
        results. However, you can only be sure to have seen all results if you
        keep calling `list` with the last `continuationToken` until you
        get a result without a `continuationToken`.

        If you are not interested in listing all the members at once, you may
        use the query-string option `limit` to return fewer.

        This method is ``experimental``
        """

        return await self._makeApiCall(self.funcinfo["listDenylist"], *args, **kwargs)

    async def heartbeat(self, *args, **kwargs):
        """
        Heartbeat

        Respond with a service heartbeat.

        This endpoint is used to check on backing services this service
        depends on.

        This method is ``stable``
        """

        return await self._makeApiCall(self.funcinfo["heartbeat"], *args, **kwargs)

    funcinfo = {
        "addDenylistAddress": {
            'args': [],
            'input': 'v1/notification-address.json#',
            'method': 'post',
            'name': 'addDenylistAddress',
            'route': '/denylist/add',
            'stability': 'experimental',
        },
        "deleteDenylistAddress": {
            'args': [],
            'input': 'v1/notification-address.json#',
            'method': 'delete',
            'name': 'deleteDenylistAddress',
            'route': '/denylist/delete',
            'stability': 'experimental',
        },
        "email": {
            'args': [],
            'input': 'v1/email-request.json#',
            'method': 'post',
            'name': 'email',
            'route': '/email',
            'stability': 'experimental',
        },
        "heartbeat": {
            'args': [],
            'method': 'get',
            'name': 'heartbeat',
            'route': '/__heartbeat__',
            'stability': 'stable',
        },
        "lbheartbeat": {
            'args': [],
            'method': 'get',
            'name': 'lbheartbeat',
            'route': '/__lbheartbeat__',
            'stability': 'stable',
        },
        "listDenylist": {
            'args': [],
            'method': 'get',
            'name': 'listDenylist',
            'output': 'v1/notification-address-list.json#',
            'query': ['continuationToken', 'limit'],
            'route': '/denylist/list',
            'stability': 'experimental',
        },
        "matrix": {
            'args': [],
            'input': 'v1/matrix-request.json#',
            'method': 'post',
            'name': 'matrix',
            'route': '/matrix',
            'stability': 'experimental',
        },
        "ping": {
            'args': [],
            'method': 'get',
            'name': 'ping',
            'route': '/ping',
            'stability': 'stable',
        },
        "pulse": {
            'args': [],
            'input': 'v1/pulse-request.json#',
            'method': 'post',
            'name': 'pulse',
            'route': '/pulse',
            'stability': 'experimental',
        },
        "slack": {
            'args': [],
            'input': 'v1/slack-request.json#',
            'method': 'post',
            'name': 'slack',
            'route': '/slack',
            'stability': 'experimental',
        },
        "version": {
            'args': [],
            'method': 'get',
            'name': 'version',
            'route': '/__version__',
            'stability': 'stable',
        },
    }


__all__ = ['createTemporaryCredentials', 'config', '_defaultConfig', 'createApiClient', 'createSession', 'Notify']



# coding=utf-8
#####################################################
# THIS FILE IS AUTOMATICALLY GENERATED. DO NOT EDIT #
#####################################################
# noqa: E128,E201
from ..client import BaseClient
from ..client import createApiClient
from ..client import config
from ..client import createTemporaryCredentials
from ..client import createSession
_defaultConfig = config


class Object(BaseClient):
    """
    The object service provides HTTP-accessible storage for large blobs of data.

    Objects can be uploaded and downloaded, with the object data flowing directly
    from the storage "backend" to the caller, and not directly via this service.
    Once uploaded, objects are immutable until their expiration time.
    """

    classOptions = {
    }
    serviceName = 'object'
    apiVersion = 'v1'

    def ping(self, *args, **kwargs):
        """
        Ping Server

        Respond without doing anything.
        This endpoint is used to check that the service is up.

        This method is ``stable``
        """

        return self._makeApiCall(self.funcinfo["ping"], *args, **kwargs)

    def lbheartbeat(self, *args, **kwargs):
        """
        Load Balancer Heartbeat

        Respond without doing anything.
        This endpoint is used to check that the service is up.

        This method is ``stable``
        """

        return self._makeApiCall(self.funcinfo["lbheartbeat"], *args, **kwargs)

    def version(self, *args, **kwargs):
        """
        Taskcluster Version

        Respond with the JSON version object.
        https://github.com/mozilla-services/Dockerflow/blob/main/docs/version_object.md

        This method is ``stable``
        """

        return self._makeApiCall(self.funcinfo["version"], *args, **kwargs)

    def createUpload(self, *args, **kwargs):
        """
        Begin upload of a new object

        Create a new object by initiating upload of its data.

        This endpoint implements negotiation of upload methods.  It can be called
        multiple times if necessary, either to propose new upload methods or to
        renew credentials for an already-agreed upload.

        The `name` parameter can contain any printable ASCII character (0x20 - 0x7e).
        The `uploadId` must be supplied by the caller, and any attempts to upload
        an object with the same name but a different `uploadId` will fail.
        Thus the first call to this method establishes the `uploadId` for the
        object, and as long as that value is kept secret, no other caller can
        upload an object of that name, regardless of scopes.  Object expiration
        cannot be changed after the initial call, either.  It is possible to call
        this method with no proposed upload methods, which has the effect of "locking
        in" the `expiration`, `projectId`, and `uploadId` properties and any
        supplied hashes.

        Unfinished uploads expire after 1 day.

        This method is ``stable``
        """

        return self._makeApiCall(self.funcinfo["createUpload"], *args, **kwargs)

    def finishUpload(self, *args, **kwargs):
        """
        Mark an upload as complete.

        This endpoint marks an upload as complete.  This indicates that all data has been
        transmitted to the backend.  After this call, no further calls to `uploadObject` are
        allowed, and downloads of the object may begin.  This method is idempotent, but will
        fail if given an incorrect uploadId for an unfinished upload.

        Note that, once `finishUpload` is complete, the object is considered immutable.

        This method is ``stable``
        """

        return self._makeApiCall(self.funcinfo["finishUpload"], *args, **kwargs)

    def startDownload(self, *args, **kwargs):
        """
        Download object data

        Start the process of downloading an object's data.  Call this endpoint with a list of acceptable
        download methods, and the server will select a method and return the corresponding payload.

        Returns a 406 error if none of the given download methods are available.

        See [Download Methods](https://docs.taskcluster.net/docs/reference/platform/object/download-methods) for more detail.

        This method is ``stable``
        """

        return self._makeApiCall(self.funcinfo["startDownload"], *args, **kwargs)

    def object(self, *args, **kwargs):
        """
        Get an object's metadata

        Get the metadata for the named object.  This metadata is not sufficient to
        get the object's content; for that use `startDownload`.

        This method is ``stable``
        """

        return self._makeApiCall(self.funcinfo["object"], *args, **kwargs)

    def download(self, *args, **kwargs):
        """
        Get an object's data

        Get the data in an object directly.  This method does not return a JSON body, but
        redirects to a location that will serve the object content directly.

        URLs for this endpoint, perhaps with attached authentication (`?bewit=..`),
        are typically used for downloads of objects by simple HTTP clients such as
        web browsers, curl, or wget.

        This method is limited by the common capabilities of HTTP, so it may not be
        the most efficient, resilient, or featureful way to retrieve an artifact.
        Situations where such functionality is required should ues the
        `startDownload` API endpoint.

        See [Simple Downloads](https://docs.taskcluster.net/docs/reference/platform/object/simple-downloads) for more detail.

        This method is ``stable``
        """

        return self._makeApiCall(self.funcinfo["download"], *args, **kwargs)

    def heartbeat(self, *args, **kwargs):
        """
        Heartbeat

        Respond with a service heartbeat.

        This endpoint is used to check on backing services this service
        depends on.

        This method is ``stable``
        """

        return self._makeApiCall(self.funcinfo["heartbeat"], *args, **kwargs)

    funcinfo = {
        "createUpload": {
            'args': ['name'],
            'input': 'v1/create-upload-request.json#',
            'method': 'put',
            'name': 'createUpload',
            'output': 'v1/create-upload-response.json#',
            'route': '/upload/<name>',
            'stability': 'stable',
        },
        "download": {
            'args': ['name'],
            'method': 'get',
            'name': 'download',
            'route': '/download/<name>',
            'stability': 'stable',
        },
        "finishUpload": {
            'args': ['name'],
            'input': 'v1/finish-upload-request.json#',
            'method': 'post',
            'name': 'finishUpload',
            'route': '/finish-upload/<name>',
            'stability': 'stable',
        },
        "heartbeat": {
            'args': [],
            'method': 'get',
            'name': 'heartbeat',
            'route': '/__heartbeat__',
            'stability': 'stable',
        },
        "lbheartbeat": {
            'args': [],
            'method': 'get',
            'name': 'lbheartbeat',
            'route': '/__lbheartbeat__',
            'stability': 'stable',
        },
        "object": {
            'args': ['name'],
            'method': 'get',
            'name': 'object',
            'output': 'v1/get-object-response.json#',
            'route': '/metadata/<name>',
            'stability': 'stable',
        },
        "ping": {
            'args': [],
            'method': 'get',
            'name': 'ping',
            'route': '/ping',
            'stability': 'stable',
        },
        "startDownload": {
            'args': ['name'],
            'input': 'v1/download-object-request.json#',
            'method': 'put',
            'name': 'startDownload',
            'output': 'v1/download-object-response.json#',
            'route': '/start-download/<name>',
            'stability': 'stable',
        },
        "version": {
            'args': [],
            'method': 'get',
            'name': 'version',
            'route': '/__version__',
            'stability': 'stable',
        },
    }


__all__ = ['createTemporaryCredentials', 'config', '_defaultConfig', 'createApiClient', 'createSession', 'Object']
